package edu.uiuc.ncsa.security.core.configuration;

import org.apache.commons.configuration.AbstractConfiguration;
import org.apache.commons.configuration.ConfigurationException;
import org.apache.commons.configuration.SubnodeConfiguration;
import org.apache.commons.configuration.XMLConfiguration;
import org.apache.commons.configuration.tree.ConfigurationNode;
import org.apache.log4j.Level;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

import java.io.File;
import java.net.URL;
import java.util.Collections;
import java.util.List;

/**
 * Utility class for getting configurations and things in them. Generally every call is
 * done anew, so anything you want to keep you should store.
 * <p>Created by Jeff Gaynor<br>
 * on 1/19/12 at  11:20 AM
 */
public class Configurations {

    /**
     * Useful utility call. This explicitly kills off logging from log 4 java. This is needed since some
     * application (don't know which) has a misconfigured log4j properties file in its library which
     * spews forth random messages. Invoke this at server start to definitely kill off any log4j activity.
     * It is annoying to have to add a dependency to clean up after someone else!
     */
    public static void killLog4J() {
        List<Logger> loggers = Collections.<Logger>list(LogManager.getCurrentLoggers());
        loggers.add(LogManager.getRootLogger());
        for (Logger logger : loggers) {
            logger.setLevel(Level.OFF);
        }
    }

    /**
     * Loads the configuration from the given file. Various checks are done to give information about
     * failures.
     * @param file
     * @return
     */
    public static XMLConfiguration getConfiguration(File file) {
        if (file == null) {
            throw new edu.uiuc.ncsa.security.core.exceptions.ConfigurationException("Error: no configuration file");
        }
        if (!file.exists()) {
            throw new edu.uiuc.ncsa.security.core.exceptions.ConfigurationException("Error: file \"" + file.getAbsolutePath() + "\" does not exist");
        }
        if (!file.isFile()) {
            throw new edu.uiuc.ncsa.security.core.exceptions.ConfigurationException("Error: \"" + file.getAbsolutePath() + "\" is not a file");
        }
        AbstractConfiguration.setDefaultListDelimiter((char) 0);
        try {
            return new XMLConfiguration(file);
        } catch (ConfigurationException e) {
            throw new edu.uiuc.ncsa.security.core.exceptions.ConfigurationException("Error: Could not create configuration from file \"" + file + "\"", e);
        }
    }

    /**
     * Gets the configuration via the resource name.
     * @param resourceName
     * @return
     */
    public static XMLConfiguration getConfiguration(URL resourceName) {
        // disable automatic list processing (which splits every value at any comma) since this plays hob with
        // values.
        try {
            AbstractConfiguration.setDefaultListDelimiter((char) 0);
            return new XMLConfiguration(resourceName);
        } catch (ConfigurationException cx) {
            throw new edu.uiuc.ncsa.security.core.exceptions.ConfigurationException("Error: could not get configuration for resource " + resourceName, cx);
        }
    }


    /**
     * Convenience method for getting the value of a single node, i.e. the contents, so
     * <br><br>
     * &lt;foo&gt;value&lt;/foo&gt;<br><br> would have name equal to 'foo' and return the
     * string 'value'.
     *
     * @param node
     * @param name
     * @return
     */
    public static String getNodeValue(ConfigurationNode node, String name) {
        List list = node.getChildren(name);
        if (list.isEmpty()) {
            return null;
        }
        return ((ConfigurationNode) list.get(0)).getValue().toString();
    }

    /**
     * Get the very first attribute with the given name from the configuration node.
     * A null is returned if there is no such value.
     * @param node
     * @param name
     * @return
     */
    public static String getFirstAttribute(ConfigurationNode node, String name) {
        List list = node.getAttributes(name);
        if (list.isEmpty()) {
            return null;
        }
        ConfigurationNode cn = (ConfigurationNode) list.get(0);
        return cn.getValue().toString();

    }

    /**
     * Return the default value if none is found.
     *
     * @param node
     * @param name
     * @param defaultValue
     * @return
     */
    public static String getNodeValue(ConfigurationNode node, String name, String defaultValue) {
        String x = getNodeValue(node, name);
        return x == null ? defaultValue : x;
    }

    /**
     * Loads the named configuration that resides under the topNodeName, e.g. topNodeName might be
     * 'config' and configName might be 'my-server'. This supports aliases too, so if the top level
     * tag contains an alias="x" attribute, then the configuration named "x" is loaded instead. This
     * allows for things like a "default" or "current" configuration which need never change once deployed.
     * The aim is to make this all as configuration driven as possible.
     * @param cfg
     * @param topNodeName
     * @param configName
     * @return
     */
    public static ConfigurationNode getConfig(XMLConfiguration cfg, String topNodeName, String configName) {
        List list = cfg.configurationsAt(topNodeName);
        ConfigurationNode configurationNode = null;
        for (int i = 0; i < list.size(); i++) {
            SubnodeConfiguration cn = (SubnodeConfiguration) list.get(i);
            List nameAttribs = cn.getRootNode().getAttributes("name");
            for (int j = 0; j < nameAttribs.size(); j++) {
                ConfigurationNode attrcn = (ConfigurationNode) nameAttribs.get(j);
                if (attrcn.getValue().equals(configName)) {
                    configurationNode = cn.getRootNode();
                }
            }
            if(configurationNode != null){
                String alias = getFirstAttribute(configurationNode, "alias");
                if(alias != null && 0 < alias.length()){
                    return getConfig(cfg, topNodeName, alias);
                }
                return configurationNode;
            }
        }
        throw new edu.uiuc.ncsa.security.core.exceptions.ConfigurationException("Configuration \"" + configName + "\" not found");

    }
}
